-- Main application class

-- TODO: Quite a BLOB here, a bit of architecturing wouldn't hurt. This is unlikely to grow though.

import "Turbine";
import "Turbine.UI";
import "Turbine.UI.Lotro";

import "Equendil.UI.WindowManager";
import "Equendil.UI.Resources";
import "Equendil.UI.IconButton";
import "Equendil.UI.ScrollBar";
import "Equendil.UI.VirtualListBox";
import "Equendil.UI.Resizer";
import "Equendil.Utils.Maths";
import "Equendil.Utils.PluginData";
import "Equendil.PluginMng.Shell";
import "Equendil.PluginMng.Plugin";
import "Equendil.PluginMng.Command";
import "Equendil.PluginMng.CommandConfigurator";
import "Equendil.PluginMng.Menu";
import "Equendil.PluginMng.MenuButton";
import "Equendil.PluginMng.KnownPlugins";

-- Default commands we always want the menu to have no matter what
local DefaultCommands = {
	-- { name = "Exit Boostrap", command = "/boostrap unload"	}, -- removed for now, no 'clean' way to exit the plugin
	{ name = "Bootstrap", command = "/bootstrap toggle", mode = Command.Mode.Always },
	-- { name = "Reload Plugins", command = "/bootstrap reload", mode = Command.Mode.Always } -- we shouldn't make that a default command really
}

-- controls displayed by the plugins listbox
local PluginItem = class( Turbine.UI.Control );
function PluginItem:Constructor( plugin )
	Turbine.UI.Control.Constructor( self );

	self.plugin = plugin;
	self.selected = false;
	
	self:SetSize( 490, 18 );
	self:SetBackColorBlendMode( Turbine.UI.BlendMode.AlphaBlend );
	
	-- auto load checkbox
	self.checkbox = Turbine.UI.Lotro.CheckBox();
	self.checkbox:SetParent( self );
	self.checkbox:SetSize( 18, 18 );
	self.checkbox:SetLeft( 10 );
	
	-- name, author, vesion labels ...
	self.labelName = Turbine.UI.Label();
	self.labelName:SetParent( self );
	self.labelName:SetParent( self );
	self.labelName:SetSize( 170, 18 );
	self.labelName:SetLeft( 40 );
	self.labelName:SetFont( Turbine.UI.Lotro.Font.Verdana14 );
	self.labelName:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
	self.labelName:SetForeColor( Equendil.UI.Color.TurbineWhite );
	self.labelName:SetMouseVisible( false );

	self.labelAuthor = Turbine.UI.Label();
	self.labelAuthor:SetParent( self );
	self.labelAuthor:SetSize( 170, 18 );
	self.labelAuthor:SetLeft( 220 );
	self.labelAuthor:SetFont( Turbine.UI.Lotro.Font.Verdana14 );
	self.labelAuthor:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
	self.labelAuthor:SetForeColor( Equendil.UI.Color.TurbineWhite );
	self.labelAuthor:SetMouseVisible( false );

	self.labelVersion = Turbine.UI.Label();
	self.labelVersion:SetParent( self );
	self.labelVersion:SetSize( 80, 18 );
	self.labelVersion:SetLeft( 400 );
	self.labelVersion:SetFont( Turbine.UI.Lotro.Font.Verdana14 );
	self.labelVersion:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
	self.labelVersion:SetForeColor( Equendil.UI.Color.TurbineWhite );
	self.labelVersion:SetMouseVisible( false );	

	-- Forward the checkbox status to the plugin
	self.checkbox.CheckedChanged = function( sender, args ) 
		self.plugin:SetAutoLoad( self.checkbox:IsChecked() );
	end
	
	-- register a callback with a plugin to update the display as needed
	self.pluginChangedAction = function() self:_Update(); end;
	self.plugin.Changed:Add( self.pluginChangedAction );
	
	self:_Update();
end

function PluginItem:_Update()
	self.checkbox:SetChecked( self.plugin:IsAutoLoad() );
	self.labelName:SetText( self.plugin:GetName() );
	self.labelAuthor:SetText( self.plugin:GetAuthor() );	
	self.labelVersion:SetText( self.plugin:GetVersion() );
	self:_SetBackColor();
	self:_SetForeColor();
	
	self.checkbox:SetEnabled( not self.plugin:IsExcluded() );
end

function PluginItem:_SetForeColor()
	local color;
	if self.plugin:IsLoaded() then
		if self.plugin:IsExcluded() then
			color = Turbine.UI.Color( 0.5, 0, 1, 0 );
		else
			color = Turbine.UI.Color( 1, 0, 1, 0 );
		end
	else
		if self.plugin:IsExcluded() then
			color = Turbine.UI.Color( 0.5, 1, 1, 1 );
		else
			color = Turbine.UI.Color( 1, 1, 1, 1 );
		end
	end

	self.labelName:SetForeColor( color );
	self.labelAuthor:SetForeColor( color );
	self.labelVersion:SetForeColor( color );
end

function PluginItem:_SetBackColor()
	if self.selected then
		self:SetBackColor( Turbine.UI.Color( 0.2, 0.9, 0.9, 1 ) );
	else
		self:SetBackColor( Turbine.UI.Color( 0, 0, 0, 0 ) );
	end
end

-- VirtualListBox can call this to set the selection status
function PluginItem:SetSelected( flag )
	self.selected = flag;
	self:_SetBackColor();
end

-- controls displayed in thec ommands listbox
local CommandItem = class( Turbine.UI.Control );
function CommandItem:Constructor( command, isDefault )
	Turbine.UI.Control.Constructor( self );

	self.command = command;
	self.isDefault = isDefault;
	
	self:SetSize( 190, 18 );
	self:SetBackColorBlendMode( Turbine.UI.BlendMode.AlphaBlend );	
	
	-- A label is really all we need to show here
	self.labelName = Turbine.UI.Label();
	self.labelName:SetParent( self );
	self.labelName:SetSize( 190, 18 );
	self.labelName:SetLeft( 0 );
	self.labelName:SetFont( Turbine.UI.Lotro.Font.Verdana14 );
	self.labelName:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
	self.labelName:SetForeColor( Equendil.UI.Color.TurbineWhite );
	self.labelName:SetMouseVisible( false );
	
	--  register a callback to update when a command changes
	self.commandChanged = function() self:_Update(); end
	self.command.Changed:Add( self.commandChanged );

	self:_Update();
end

function CommandItem:_Update()
	self.labelName:SetText( self.command:GetName() );
end

function CommandItem:SetSelected( flag )
	if flag then
		self:SetBackColor( Turbine.UI.Color( 0.2, 0.9, 0.9, 1 ) );
	else
		self:SetBackColor( Turbine.UI.Color( 0, 0, 0, 0 ) );
	end
end

-- Main application class
App = class( Turbine.UI.Lotro.Window );
function App:Constructor()
	Turbine.UI.Lotro.Window.Constructor( self );
	
	-- Variables

	self.plugins = {}; -- plugins
	self.pluginItems = {}; -- plugins listbox elements, indexed by plugins
	self.commands = {}; -- commands
	self.commandItems = {}; -- commands listbox elements, indexed by commands

	self.includedPlugins = {}; -- indirection table for plugins that are not in the exclusion list. Values are indices of self.plugins
	
	self.persistentData = {}; -- stored data for plugins & commands

	self.version = Plugins["Bootstrap"]:GetVersion(); -- version of this plugin

	self.showAllPlugins = false;
	
	-- UI elements
	
	-- setup a window and center it
    self:SetSize( 575, 600 );
	self:SetPosition( 
		( Turbine.UI.Display:GetWidth() - self:GetWidth() ) / 2, 
		( Turbine.UI.Display:GetHeight() - self:GetHeight() ) /2 
	);	
	self:SetText( "Bootstrap - Plugin Manager" );
	self:SetParent( nil );
	Equendil.UI.WindowManager:ManageWindow( self ); -- makes this window managed for HUD toggle / ESC
	
	self.resizer = Equendil.UI.Resizer();
	self.resizer:AttachWindow( self, 575, 600, 575, Turbine.UI.Display.GetHeight() );
	
	-- Add a background so it does not look too dull
	self.background = Equendil.UI.DecoratedContainer( Equendil.UI.Decoration.GrainyBlue );
	self.background:SetParent( self );
	self.background:SetSize( 512, 560 );
	self.background:SetPosition( 31, 30 );

	-- PLUGINS ELEMENTS
	
	-- Container for all controls related to plugins
	self.backgroundPlugins = Turbine.UI.Control();
	self.backgroundPlugins:SetParent( self );
	self.backgroundPlugins:SetSize( 530, 270 );
	self.backgroundPlugins:SetPosition( 22, 30 );
	
	-- listbox headers
	self.labelPluginsHeaderAutoLoad = Equendil.UI.CommonControl.LabelListHeading( "Auto" );
	self.labelPluginsHeaderAutoLoad:SetParent( self.backgroundPlugins );
	self.labelPluginsHeaderAutoLoad:SetSize( 40, 14 );
	self.labelPluginsHeaderAutoLoad:SetPosition( 10, 5 );
	
	self.labelPluginsHeaderName = Equendil.UI.CommonControl.LabelListHeading( "Plugin Name" );
	self.labelPluginsHeaderName:SetParent( self.backgroundPlugins );
	self.labelPluginsHeaderName:SetSize( 170, 14 );
	self.labelPluginsHeaderName:SetPosition( 50, 5 );
	
	self.labelPluginsHeaderAuthor = Equendil.UI.CommonControl.LabelListHeading( "Author" );
	self.labelPluginsHeaderAuthor:SetParent( self.backgroundPlugins );
	self.labelPluginsHeaderAuthor:SetSize( 170, 14 );
	self.labelPluginsHeaderAuthor:SetPosition( 230, 5 );

	self.labelPluginsHeaderVersion = Equendil.UI.CommonControl.LabelListHeading( "Version" );
	self.labelPluginsHeaderVersion:SetParent( self.backgroundPlugins );
	self.labelPluginsHeaderVersion:SetSize( 70, 14 );
	self.labelPluginsHeaderVersion:SetPosition( 410, 5 );

	-- borders around the plugins listbox
	self.outlinePlugins = Equendil.UI.DecoratedContainer( Equendil.UI.Decoration.RoundSilverBlue );
	self.outlinePlugins:SetParent( self.backgroundPlugins );
	self.outlinePlugins:SetSize( 520, 220 );
	self.outlinePlugins:SetPosition( 0, 20 );

	-- scrollbar for listbox
	self.scrollbarPlugins = Equendil.UI.ScrollBar( Turbine.UI.Orientation.Vertical );
	self.scrollbarPlugins:SetParent( self.backgroundPlugins );
	self.scrollbarPlugins:SetSize( 10, 200 );
	self.scrollbarPlugins:SetPosition( 500, 30 );

	-- Plugins listbox
	local getPluginItem = function( n, selection )
		local pluginItem = self:_GetNthPluginItem(n);
		pluginItem:_Update();
		pluginItem:SetSelected( selection );
		return pluginItem;
	end
	self.listboxPlugins = Equendil.UI.VirtualListBox( getPluginItem, 18 );
	self.listboxPlugins:SetParent( self.backgroundPlugins );
	self.listboxPlugins:SetSize( 490, 200 );
	self.listboxPlugins:SetPosition( 10, 30 );
	-- self.listboxPlugins:SetItemHeight( 18 );
	self.listboxPlugins:SetVerticalScrollBar( self.scrollbarPlugins );
	self.listboxPlugins:SetSelectionMode( Equendil.UI.SelectionMode.SingleSelection );
	self.listboxPlugins:SetBackColor( Turbine.UI.Color( 0.1, 1, 1, 1 ) );
	self.listboxPlugins:SetBackColorBlendMode( Turbine.UI.BlendMode.Overlay );

	-- buttons for actions on plugins
	
	-- up/down buttons to order plugins
	self.buttonPluginUp = Equendil.UI.ButtonArrowUp20();
	self.buttonPluginUp:SetParent( self.backgroundPlugins );
	self.buttonPluginUp:SetPosition( 10, 243 );
	
	self.buttonPluginDown = Equendil.UI.ButtonArrowDown20();
	self.buttonPluginDown:SetParent( self.backgroundPlugins );
	self.buttonPluginDown:SetPosition( 30, 243 );

	-- "Load plugin" action
	self.buttonPluginLoad = Turbine.UI.Lotro.Button();
	self.buttonPluginLoad:SetParent( self.backgroundPlugins );
	self.buttonPluginLoad:SetSize( 80, 20 );
	self.buttonPluginLoad:SetPosition( 60, 243 );
	self.buttonPluginLoad:SetText( "Launch" );
	self.buttonPluginLoad:SetEnabled( false );

	self.buttonPluginVisibility = Turbine.UI.Lotro.Button();
	self.buttonPluginVisibility:SetParent( self.backgroundPlugins );
	self.buttonPluginVisibility:SetSize( 80, 20 );
	self.buttonPluginVisibility:SetPosition( 150, 243 );
	self.buttonPluginVisibility:SetText( "Hide" );
	self.buttonPluginVisibility:SetEnabled( false );
	
	-- "Add to Menu" action
	self.buttonPluginsAddCommand = Turbine.UI.Lotro.Button();
	self.buttonPluginsAddCommand:SetParent( self.backgroundPlugins );
	self.buttonPluginsAddCommand:SetSize( 120, 20 );
	self.buttonPluginsAddCommand:SetPosition( 240, 243 );
	self.buttonPluginsAddCommand:SetText( "Add to Menu" );
	self.buttonPluginsAddCommand:SetEnabled( false );
	
	-- show hidden plugins check box & label
	self.checkboxShowHiddenPlugins = Turbine.UI.Lotro.CheckBox();
	self.checkboxShowHiddenPlugins:SetParent( self.backgroundPlugins );
	self.checkboxShowHiddenPlugins:SetSize( 18, 18 );
	self.checkboxShowHiddenPlugins:SetPosition( 370, 243 );
	self.checkboxShowHiddenPlugins:SetChecked( self.showAllPlugins );

	self.labelShowHiddenPlugins = Equendil.UI.CommonControl.LabelFieldHeading( "Show all" );
	self.labelShowHiddenPlugins:SetParent( self.backgroundPlugins );
	self.labelShowHiddenPlugins:SetSize( 110, 18 );
	self.labelShowHiddenPlugins:SetPosition( 390, 243 );
		
	
	--[[
	self.buttonPluginsUnload = Turbine.UI.Lotro.Button();
	self.buttonPluginsUnload:SetParent( self.backgroundPlugins );
	self.buttonPluginsUnload:SetSize( 120, 20 );
	self.buttonPluginsUnload:SetPosition( 270, 243 );
	self.buttonPluginsUnload:SetText( "Unload" );
	self.buttonPluginsUnload.Click = function( sender, args )
		local n = self.listboxPlugins:GetSelectedIndex();
		if n ~= -1 then
			self.plugins[n].plugin:Unload();
		end
	end--]]

	-- COMMANDS 
	
	-- container for all controls related to commands
	self.backgroundCommands = Turbine.UI.Control();
	self.backgroundCommands:SetParent( self );
	self.backgroundCommands:SetSize( 530, 230 );
	self.backgroundCommands:SetPosition( 22, 310 );
	
	-- headers
	self.labelCommandsHeader = Equendil.UI.CommonControl.LabelListHeading( "Menu Commands" );
	self.labelCommandsHeader:SetParent( self.backgroundCommands );
	self.labelCommandsHeader:SetSize( 200, 14 );
	self.labelCommandsHeader:SetPosition( 0, 5 );
	
	self.labelCommandsHeaderDetail = Equendil.UI.CommonControl.LabelListHeading( "Detail" );
	self.labelCommandsHeaderDetail:SetParent( self.backgroundCommands );
	self.labelCommandsHeaderDetail:SetSize( 270, 14 );
	self.labelCommandsHeaderDetail:SetPosition( 230, 5 );
	
	-- borders around the listbox
	self.outlineCommands = Equendil.UI.DecoratedContainer( Equendil.UI.Decoration.RoundSilverBlue );
	self.outlineCommands:SetParent( self.backgroundCommands );
	self.outlineCommands:SetSize( 220, 182 );
	self.outlineCommands:SetPosition( 0, 20 );

	-- scrollbar for the list box
	self.scrollbarCommands = Equendil.UI.ScrollBar( Turbine.UI.Orientation.Vertical );
	self.scrollbarCommands:SetParent( self.backgroundCommands );
	self.scrollbarCommands:SetSize( 10, 162 );
	self.scrollbarCommands:SetPosition( 200, 30 );	

	-- Listbox for commands
	local getCommandItem = function( n, selection )
		local commandItem = self.commandItems[self.commands[n]];
		commandItem:SetSelected( selection );
		return commandItem;
	end
	self.listboxCommands = Equendil.UI.VirtualListBox( getCommandItem, 18 );
	self.listboxCommands:SetParent( self.backgroundCommands );
	self.listboxCommands:SetSize( 190, 162 );
	self.listboxCommands:SetPosition( 10, 30 );
	-- self.listboxCommands:SetItemHeight( 18 );
	self.listboxCommands:SetVerticalScrollBar( self.scrollbarCommands );
	self.listboxCommands:SetSelectionMode( Equendil.UI.SelectionMode.SingleSelection );
	self.listboxCommands:SetBackColor( Turbine.UI.Color( 0.1, 1, 1, 1 ) );
	self.listboxCommands:SetBackColorBlendMode( Turbine.UI.BlendMode.Overlay );
		
	-- up/down buttons to order commands within the menu
	self.buttonCommandUp = Equendil.UI.ButtonArrowUp20();
	self.buttonCommandUp:SetParent( self.backgroundCommands );
	self.buttonCommandUp:SetPosition( 0, 205 );
	
	self.buttonCommandDown = Equendil.UI.ButtonArrowDown20();
	self.buttonCommandDown:SetParent( self.backgroundCommands );
	self.buttonCommandDown:SetPosition( 20, 205 );
	
	-- "Remove command" action
	self.buttonCommandRemove = Turbine.UI.Lotro.Button();
	self.buttonCommandRemove:SetParent( self.backgroundCommands );
	self.buttonCommandRemove:SetSize( 80, 20 );
	self.buttonCommandRemove:SetPosition( 40, 205 );
	self.buttonCommandRemove:SetText( "Remove" );
	
	-- "Save Menu" action
	self.buttonCommandsSave = Turbine.UI.Lotro.Button();
	self.buttonCommandsSave:SetParent( self.backgroundCommands );
	self.buttonCommandsSave:SetSize( 100, 20 );
	self.buttonCommandsSave:SetPosition( 130, 205 );
	self.buttonCommandsSave:SetText( "Save Menu" );
	
	-- borders & external controls to fill in details about a given command
	self.outlineCommandConfigurator = Equendil.UI.DecoratedContainer( Equendil.UI.Decoration.RoundSilverBlue );
	self.outlineCommandConfigurator:SetParent( self.backgroundCommands );
	self.outlineCommandConfigurator:SetSize( 300, 182 );
	self.outlineCommandConfigurator:SetPosition( 220, 20 );
	
	self.commandConfigurator = Equendil.PluginMng.CommandConfigurator();
	self.commandConfigurator:SetParent( self.backgroundCommands );
	self.commandConfigurator:SetSize( 280, 162 );
	self.commandConfigurator:SetPosition( 230, 30 );
	self.commandConfigurator:SetBackColor( Turbine.UI.Color( 0.1, 0.5, 1, 0.5 ) );
	self.commandConfigurator:SetBackColorBlendMode( Turbine.UI.BlendMode.AlphaBlend );

	-- menu button opacity labels & scrollbar
	self.labelMenuButtonOpacity = Equendil.UI.CommonControl.LabelFieldHeading( "Button Opacity" );
	self.labelMenuButtonOpacity:SetParent( self.backgroundCommands );
	self.labelMenuButtonOpacity:SetSize( 120, 18 );
	self.labelMenuButtonOpacity:SetPosition( 250, 205 );
	
	self.scrollBarMenuButtonOpacity = Equendil.UI.ScrollBar( Turbine.UI.Orientation.Horizontal );
	self.scrollBarMenuButtonOpacity:SetParent( self.backgroundCommands );
	self.scrollBarMenuButtonOpacity:SetSize( 100, 10 );
	self.scrollBarMenuButtonOpacity:SetPosition( 380, 209 );
	self.scrollBarMenuButtonOpacity:SetLargeChange( 10 );
	self.scrollBarMenuButtonOpacity:SetMinimum( 0 );
	self.scrollBarMenuButtonOpacity:SetMaximum( 100 );
	self.scrollBarMenuButtonOpacity:SetSliderProportion( 0.2 );

	self.labelMenuButtonOpacityValue = Turbine.UI.Label();
	self.labelMenuButtonOpacityValue:SetParent( self.backgroundCommands );
	self.labelMenuButtonOpacityValue:SetFont( Turbine.UI.Lotro.Font.TrajanPro16 );
	self.labelMenuButtonOpacityValue:SetTextAlignment( Turbine.UI.ContentAlignment.MiddleLeft );
	self.labelMenuButtonOpacityValue:SetForeColor( Equendil.UI.Color.TurbineWhite );
	self.labelMenuButtonOpacityValue:SetFontStyle( Turbine.UI.FontStyle.Outline );
	self.labelMenuButtonOpacityValue:SetSize( 40, 18 );
	self.labelMenuButtonOpacityValue:SetPosition( 490, 205 );
	self.labelMenuButtonOpacityValue:SetMouseVisible( false );
	
	-- RELOAD BUTTON

	-- "Refresh List" button
	self.buttonPluginsRefresh = Turbine.UI.Lotro.Button();
	self.buttonPluginsRefresh:SetParent( self );
	self.buttonPluginsRefresh:SetPosition( 127, 560 );
	self.buttonPluginsRefresh:SetSize( 150, 20 );
	self.buttonPluginsRefresh:SetText( "Refresh Plugins" );

	
	self.buttonReload = Turbine.UI.Lotro.Button();
	self.buttonReload:SetParent( self );
	self.buttonReload:SetPosition( 297, 560 );
	self.buttonReload:SetSize( 150, 20 );
	self.buttonReload:SetText( "Reload Bootstrap" );
	
	-- MENU / START BUTTON
	self.menu = Equendil.PluginMng.Menu();
	
	self.menuButton = Equendil.PluginMng.MenuButton( self.menu );
	self.menuButton:SetParent( nil );
	self.menuButton:SetVisible( true );

	-- kill the reloader if needed
	self:_KillReloader();

	-- load data
	self:Load();

	-- update scrollbar & label for opacity
	self.labelMenuButtonOpacityValue:SetText( tostring( self.menuButton:GetOpacity() * 100 ) .. "%" );
	self.scrollBarMenuButtonOpacity:SetValue( self.menuButton:GetOpacity() * 100 );
	
	-- ACTIONS

	-- action on plugins 
	self.listboxPlugins.SelectionChanged = function() self:_PluginsSelectionChanged() end
	self.buttonPluginUp.Clicked = function( button ) self:_PluginUp() end
	self.buttonPluginDown.Clicked = function( button ) self:_PluginDown() end
	self.buttonPluginsAddCommand.Click = function( sender, args ) self:_AddCommand() end
	self.buttonPluginLoad.Click = function( sender, args ) self:_LoadSelectedPlugin() end
	self.buttonPluginVisibility.Click = function( sender, args ) self:_TogglePluginVisibility() end
	self.checkboxShowHiddenPlugins.CheckedChanged = function( sender, args ) self:ShowAllPlugins( self.checkboxShowHiddenPlugins:IsChecked() ) end

	-- action on commands
	self.listboxCommands.SelectionChanged = function( sender, args ) self:_CommandsSelectionChanged() end
	self.buttonCommandUp.Clicked = function( button ) self:_CommandUp() end
	self.buttonCommandDown.Clicked = function( button ) self:_CommandDown() end
	self.buttonCommandRemove.Click = function( sender, args ) self:_RemoveSelectedCommand() end
	self.buttonCommandsSave.Click = function( sender, args ) self:SaveCommands(); end

	-- show menu option
	-- self.checkboxMenuButtonVisible.CheckedChanged = function( sender, args ) self:_SetMenuButtonVisibility() end;
	self.scrollBarMenuButtonOpacity.ValueChanged = function() self:_SetMenuButtonOpacity() end

	
	-- Global actions
	self.buttonReload.Click = function( sender, args ) self:Reload(); end
	self.buttonPluginsRefresh.Click = function( sender, args ) self:RefreshPlugins(); end	
	
	-- action for use whenever a plugin is set to autoload
	self.actionSaveAutoLoad = function() self:SaveAutoLoad(); end	

	
	-- now that window & button should be in position, register callbacks for saving those if they move/change.
	self.menuButton.Changed = function() self:SaveMenu(); end
	self.Closed = function( sender, args ) self:SaveWindow(); end
	self.PositionChanged = function( sender, args ) self:SaveWindow(); end
	self.resizer.Resized = function() self:SaveWindow(); end
	
	-- create a Shell object so we can execute plugin commands
	self.shell = Equendil.PluginMng.Shell();
	self.shell:Hook(); -- hook Turbine.Shell to catch other plugins' commands

	-- register a function to unhook the shell when this plugin is about to exit
	Plugins["Bootstrap"].Unload = function()
		self.shell:Unhook(); -- we really need to clear our hook before we go
	end	
	
	-- Finally, build the list of plugins (and commands)
	self:RefreshPlugins();
	
	-- Output something in the chat so the user knows the plugin has started
	Turbine.Shell.WriteLine( "Boostrap version " .. self.version .. " started." );
	
	-- Load all plugins set to auto
	self:AutoLoadPlugins();
end

function App:SetSize( width, height )
	Turbine.UI.Window.SetSize( self, width, height );
	self:_DoLayout();
end

function App:SetWidth( width )
	Turbine.UI.Window.SetWidth( self, width );
	self:_DoLayout();
end

function App:SetHeight( height )
	Turbine.UI.Window.SetHeight( self, height );
	self:_DoLayout();
end

function App:_DoLayout()
	local width, height = self:GetSize();

	if self.background == nil then return end -- return if not initialized yet
	
	self.background:SetHeight( height - 40 );
	
	local additionalHeight = ( height - 600 ) / 2;
	self.backgroundPlugins:SetHeight( 530, 270 + additionalHeight );
	self.outlinePlugins:SetHeight( 220 + additionalHeight );
	self.scrollbarPlugins:SetHeight( 200 + additionalHeight );
	self.listboxPlugins:SetHeight( 200 + additionalHeight );
	self.buttonPluginUp:SetTop( 243 + additionalHeight );
	self.buttonPluginDown:SetTop( 243 + additionalHeight );
	self.buttonPluginLoad:SetTop( 243 + additionalHeight );
	self.buttonPluginVisibility:SetTop( 243 + additionalHeight );
	self.buttonPluginsAddCommand:SetTop( 243 + additionalHeight );
	self.checkboxShowHiddenPlugins:SetTop( 243 + additionalHeight );
	self.labelShowHiddenPlugins:SetTop( 243 + additionalHeight );
	
	self.backgroundCommands:SetTop( 310 + additionalHeight );
	self.backgroundCommands:SetHeight( 230 + additionalHeight );
	self.outlineCommands:SetHeight( 182 + additionalHeight );
	self.scrollbarCommands:SetHeight( 162 + additionalHeight );
	self.listboxCommands:SetHeight( 162 + additionalHeight );
	self.buttonCommandUp:SetTop( 205 + additionalHeight );
	self.buttonCommandDown:SetTop( 205 + additionalHeight );
	self.buttonCommandRemove:SetTop( 205 + additionalHeight );
	self.buttonCommandsSave:SetTop( 205 + additionalHeight );
	self.outlineCommandConfigurator:SetHeight( 182 + additionalHeight );
	self.commandConfigurator:SetHeight( 162 + additionalHeight );
	self.labelMenuButtonOpacity:SetTop( 205 + additionalHeight );
	self.scrollBarMenuButtonOpacity:SetTop( 209 + additionalHeight );
	self.labelMenuButtonOpacityValue:SetTop(  205 + additionalHeight );
	
	self.buttonPluginsRefresh:SetTop( 560 + additionalHeight * 2 );
	self.buttonReload:SetTop( 560 + additionalHeight * 2 );
end			

-- show/hide the window
function App:_Show( flag )
	self:SetVisible( flag );
	if flag then
		self:Activate();
		self.listboxPlugins:Refresh(); -- refresh the listbox to check for loaded states 
	end
end

-- external method of the same
function App:Show( flag )
	self:_Show( flag );
	self:SaveWindow();
end

function App:Toggle()
	self:Show( not self:IsVisible() );
end

-- Nuke all plugins including self
function App:Nuke()
	-- TODO: no way to do it properly atm
end

-- Kill the reloader if it's loaded
function App:_KillReloader()
	-- If BootstrapReloader called us, UnloadScriptState() called immediately will fail, so
	-- we defer the call using an Update event.
	self.Update = function() 
		Turbine.PluginManager.UnloadScriptState( "BootstrapReloader" );
		self:SetWantsUpdates( false );
		self.Update = nil;
	end
	self:SetWantsUpdates( true );
end

function App:Reload()
	self.Closed = nil; -- window will be closed if open, not a user action, we do *not* want to know about it

	-- the reloader will kill everything and restart bootstrap
	Turbine.PluginManager.LoadPlugin( "BootstrapReloader" );
end

-- equivalent to /plugins refresh + /plugins list in turbine's shell
function App:RefreshPlugins()
	Turbine.PluginManager.RefreshAvailablePlugins();
	self.commandConfigurator:SetCommand( nil );
	self:_FetchPlugins();
	self:_FetchCommands();
end

-- Load all plugins set to "auto" load
function App:AutoLoadPlugins()
	local outputString = "Bootstrap: Auto-loaded ";
	local separator = "";
	for i, plugin in ipairs( self.plugins ) do
		if plugin:IsAutoLoad() then
			outputString = outputString .. separator .. plugin:GetName();
			if separator == "" then separator = ", " end
			plugin:Load();
		end
	end
	Turbine.Shell.WriteLine( outputString );
end

function App:ShowAllPlugins( flag )
	self.showAllPlugins = flag;
	self:SavePlugins();
	if not flag then
		self:_BuildInclusionTable();
	end
	self:_RefreshPluginsListBox();
end

function App:_GetNthPluginItem( n )
	if self.showAllPlugins then
		return self.pluginItems[self.plugins[n]];
	else
		return self.pluginItems[self.plugins[self.includedPlugins[n]]];
	end
end

function App:_RefreshPluginsListBox()
	-- refresh the list of plugins
	local itemCount;
	if self.showAllPlugins then
		itemCount = #self.plugins;
	else
		itemCount = #self.includedPlugins;
	end
		
	self.listboxPlugins:SetItemCount( itemCount );
	self.listboxPlugins:Refresh();
end

function App:_GetSelectedPlugin()
	local n = self.listboxPlugins:GetSelectedIndex();
	if n == -1 then return nil end
	
	if self.showAllPlugins then
		return self.plugins[n];
	else
		return self.plugins[self.includedPlugins[n]];
	end
end

-- handle selection of a plugin
function App:_PluginsSelectionChanged()
	local plugin = self:_GetSelectedPlugin();
	if plugin ~= nil then
		if plugin:IsExcluded() then
			self.buttonPluginVisibility:SetText( "Show" );
		else
			self.buttonPluginVisibility:SetText( "Hide" );
		end
	end
	
	self.buttonPluginVisibility:SetEnabled( plugin ~= nil );
	self.buttonPluginLoad:SetEnabled( plugin ~= nil );
	self.buttonPluginsAddCommand:SetEnabled( plugin ~= nil );
end
	
-- move a plugin up the list of plugins	
function App:_PluginUp()
	local n = self.listboxPlugins:GetSelectedIndex();
	if n > 1 then -- check we have something selected and it's not top of the list already
		-- find indices to swap depending on whether we are showing all plugins or not
		local n1, n2 = n-1, n;
		if not self.showAllPlugins then
			-- we don't show all plugins, use the includedPlugins table for indirection
			n1, n2 = self.includedPlugins[n1], self.includedPlugins[n2];
		end
		
		-- alter plugins order and refresh the listbox
		self.plugins[n1], self.plugins[n2] = self.plugins[n2], self.plugins[n1]; -- swap
		self.listboxPlugins:Select( n - 1 );
		self.listboxPlugins:Refresh();
		self.listboxPlugins:MakeVisible( n - 1);
		
		self:SavePlugins();
	end
end

-- move a plugin down the list of plugins	
function App:_PluginDown()
	local n = self.listboxPlugins:GetSelectedIndex();
	if n ~= -1 and n ~= self.listboxPlugins:GetItemCount() then -- check if we have something selected and it's not last of list already
		-- find indices to swap depending on whether we are showing all plugins or not
		local n1, n2 = n+1, n;
		if not self.showAllPlugins then
			n1, n2 = self.includedPlugins[n1], self.includedPlugins[n2];
		end
		
		-- alter plugins order and refresh the listbox
		self.plugins[n1], self.plugins[n2] = self.plugins[n2], self.plugins[n1]; -- swap
		self.listboxPlugins:Select( n + 1 );
		self.listboxPlugins:Refresh();
		self.listboxPlugins:MakeVisible( n + 1 );
		
		self:SavePlugins(); -- we save the list of plugins automatically
	end
end

-- load the currently selected plugin
function App:_LoadSelectedPlugin()
	local plugin = self:_GetSelectedPlugin();
	if plugin ~= nil then
		plugin:Load();
	end
end

function App:_TogglePluginVisibility()
	local plugin = self:_GetSelectedPlugin();
	if plugin ~= nil then
		plugin:SetExcluded( not plugin:IsExcluded() );
	end
	if not self.showAllPlugins then
		self:_BuildInclusionTable();
	end
	
	self:_RefreshPluginsListBox();
	self:SavePlugins();
end

-- return index & PluginItem object in he list of plugins from its name. (-1, nil) if not found
function App:_FindPlugin( name )
	for i, plugin in ipairs( self.plugins ) do
		if plugin:GetName() == name then
			return i, plugin;
		end
	end
	
	return -1, nil;
end

-- returns whether a plugin is in the auto load table
function App:_IsAutoLoad( name )
	if self.persistentData == nil or self.persistentData.autoLoad == nil then
		return false;
	end
	
	for i, autoLoadPluginName in ipairs( self.persistentData.autoLoad ) do
		if autoLoadPluginName == name then
			return true;
		end
	end
end

-- returns whether a plugin is in the exclusion list of 'known' plugins
function App:_IsKnownPluginExcluded( pluginName )
	for k, excludedName in pairs( KnownPlugins.ExcludedPlugins ) do
		if pluginName == excludedName then
			return true; -- found plugin -> excluded
		end
	end
	
	return false; -- not found = not excluded
end

-- Make and initialize a new plugin
-- isExcluded can be true, false or nil, if nil, the exclusion status of the plugin will be fetched from the 'known' plugins list
function App:_NewPlugin( turbinePlugin, isExcluded )
	local plugin = Plugin( turbinePlugin );
	
	if isExcluded == nil then
		-- no exclusion known, check if it's a 'known' excluded plugin
		isExcluded = self:_IsKnownPluginExcluded( plugin:GetName() );
	end
	plugin:SetExcluded( isExcluded );
	
	-- Register a callback if the plugin for saving autoload changes if the plugin is not in the exclusion list
	-- if it is, we make double sure it won't load automatically
	if isExcluded then
		plugin:SetAutoLoad( false );
	else
		plugin:SetAutoLoad( self:_IsAutoLoad( plugin:GetName() ) );
		plugin.AutoLoadChanged:Add( self.actionSaveAutoLoad );
	end
	
	self.pluginItems[plugin] = PluginItem( plugin );
	
	return plugin;
end

-- Builds the 'includsion' table, ie, a table of plugins that are not 'excluded'
function App:_BuildInclusionTable()
	self.includedPlugins = {};
	for i, plugin in ipairs( self.plugins ) do
		if not plugin:IsExcluded() then
			table.insert( self.includedPlugins, i );
		end
	end
end

-- get the list of existing plugins, and set data of associated commands
function App:_FetchPlugins()
	-- reset the plugin table
	for _, plugin in ipairs( self.plugins ) do
		plugin.AutoLoadChanged:Remove( self.ActionSaveAutoLoad );
	end
	self.plugins = {};
	self.pluginItems = {};
	
	-- get the list of live plugins
	local turbinePlugins = Turbine.PluginManager.GetAvailablePlugins();

	-- make a temporary copy of the table indexed by plugin names for lookup
	local turbinePluginsByName = {};
	for _, turbinePlugin in pairs( turbinePlugins ) do
		turbinePluginsByName[turbinePlugin.Name] = turbinePlugin;
	end

	-- add stored plugins first in order
	if self.persistentData ~= nil and self.persistentData.plugins ~= nil then
		for _, storedPlugin in ipairs( self.persistentData.plugins ) do
			local turbinePlugin = turbinePluginsByName[storedPlugin.name];
			if turbinePlugin ~= nil then
				local plugin = self:_NewPlugin( turbinePlugin, storedPlugin.excluded );
				turbinePluginsByName[storedPlugin.name] = nil; -- remove plugin from our lookup table
				table.insert( self.plugins, plugin );
			end
		end
	end
	
	-- add remaining 'unknown' plugins in undefined order
	for _, turbinePlugin in pairs( turbinePluginsByName ) do
		local plugin = self:_NewPlugin( turbinePlugin );
		table.insert( self.plugins, plugin );
	end
	
	self:_BuildInclusionTable();
	self:_RefreshPluginsListBox();
end

-- handle selection of a command
function App:_CommandsSelectionChanged()
	local n = self.listboxCommands:GetSelectedIndex();
	self.buttonCommandRemove:SetEnabled( n > #DefaultCommands );
	if n ~= -1 then -- check we have something selected
		-- setup the external control with that command
		local command = self.commands[n];
		self.commandConfigurator:SetCommand( command );
		self.commandConfigurator:SetReadOnly( self.commandItems[command].isDefault );
	end
end

-- move a command up the list of commands
function App:_CommandUp()
	local n = self.listboxCommands:GetSelectedIndex();
	if n > #DefaultCommands + 1 then -- make sure we have something selected and it's not a default command (they stay at the top)
		self.commands[n-1], self.commands[n] = self.commands[n], self.commands[n-1]; -- swap
		-- self.commandItems[n-1], self.commandItems[n] = self.commandItems[n], self.commandItems[n-1]; -- swap 
		self.listboxCommands:Select( n - 1 );
		self.listboxCommands:Refresh();
		self.listboxCommands:MakeVisible( n - 1 );
		
		self.menu:Set( self.commands ); -- feed the 'new' command list to the menu again
	end
end

-- move a command down the list of commands
function App:_CommandDown()
	local n = self.listboxCommands:GetSelectedIndex();
	if n > #DefaultCommands and n ~= self.listboxCommands:GetItemCount() then -- make sure we have something selected, not at the bottom already, not a default command (they stay at the top)
		self.commands[n+1], self.commands[n] = self.commands[n], self.commands[n+1 ];
		-- self.commandItems[n+1], self.commandItems[n] = self.commandItems[n], self.commandItems[n + 1 ]; -- swap
		self.listboxCommands:Select( n + 1 );
		self.listboxCommands:Refresh();
		self.listboxCommands:MakeVisible( n + 1 );
		
		self.menu:Set( self.commands ); -- feed the 'new' command list to the menu again
	end
end

-- remove the selected command from the menu
function App:_RemoveSelectedCommand()
	local n = self.listboxCommands:GetSelectedIndex();
	if n > #DefaultCommands then -- make sure we have something selected and it's not a default command
		self.commandItems[self.commands[n]] = nil;
		table.remove( self.commands, n )
		self.listboxCommands:SetItemCount( #self.commands );
		self.listboxCommands:Refresh();
		
		self.menu:Set( self.commands ); -- feed the 'new' command list to the menu
	end
end	

-- set the menu button visibility according to the checkbox status
function App:_SetMenuButtonOpacity()
	self.labelMenuButtonOpacityValue:SetText( tostring( self.scrollBarMenuButtonOpacity:GetValue() ) .. "%" );
	self.menuButton:SetOpacity( self.scrollBarMenuButtonOpacity:GetValue() / 100 );
	-- self:SaveMenu(); -- will be done in menuButton.Changed()
end

-- create the list of commands
function App:_FetchCommands()
	-- create commands
	self:_CreateDefaultCommands();
	self:_AddSavedCommands();
	
	-- refresh the commands listbox & menu
	self.listboxCommands:SetItemCount( #self.commands );
	self.listboxCommands:Refresh();
	self.menu:Set( self.commands );
end

-- add commands that we have stored
function App:_AddSavedCommands()
	if self.persistentData == nil or self.persistentData.commands == nil then
		return;
	end

	for k, commandData in ipairs( self.persistentData.commands ) do
		local i, plugin = self:_FindPlugin( commandData.pluginName );
		if plugin ~= nil then
			-- found that plugin, add command
			local command = Command( self.shell );
			command:SetPlugin( plugin );
			command:SetName( commandData.name );
			command:SetShellCommand( commandData.shellCommand );
			command:SetMode( commandData.mode );
			table.insert( self.commands, command );
			self.commandItems[command] = CommandItem( command, false );
			self.menu:Set( self.commands );
		end
	end
end

-- create a command from a template. plugin = plugin to attach to the command
function App:_CreateCommandFromTemplate( template, plugin )
	local command = Command( self.shell );
	command:SetName( template.name );
	command:SetMode( template.mode );
	command:SetShellCommand( template.command );
	command:SetPlugin( plugin );
	return command;
end

-- reset commands and create a bunch of commands to use by default
function App:_CreateDefaultCommands()
	-- remove callbacks & clear tables
	for k, v in pairs( self.commands ) do 
		v.Changed:Remove( self.commandChanged );
	end

	self.commands = {};
	self.commandItems = {};

	-- create commands from default templates
	for i, commandTemplate in ipairs( DefaultCommands ) do
		local command = self:_CreateCommandFromTemplate( commandTemplate, nil );
		self.commands[i] = command;
		
		self.commandItems[command] = CommandItem( command, true );
	end
end

-- Find a template for a given plugin name if any
function App:_FindKnownPluginsCommandTemplate( pluginName )
	for k, template in pairs( KnownPlugins.CommandTemplates ) do
		if k == pluginName then
			return template;
		end
	end
	
	return nil;
end

-- add a new command for the selected plugin. If we have a template for it, we use that, if not, we setup a default command
function App:_AddCommand()
	local plugin = self:_GetSelectedPlugin();
	if plugin == nil then return end
	
	local command = nil;

	-- we're 'smart'  here, if we have a default command template for this plugin, add that, otherwise build a dummy one
	local template = self:_FindKnownPluginsCommandTemplate( plugin:GetName() );
	if template ~= nil then
		-- we've got a default command template for this plugin, add it
		command = self:_CreateCommandFromTemplate( template, plugin );
	else
		command = Equendil.PluginMng.Command( self.shell );
		command:SetShellCommand( "/do something" );
		command:SetName( plugin:GetName() );
		command:SetPlugin( plugin );
		if plugin:GetScriptState() ~= "Default" then
			command:SetMode( Command.Mode.LoadOnly ); -- non default script state, can't run command on this plugin
		else
			command:SetMode( Command.Mode.LoadOrRun );
		end
	end
	
	-- add the new command to out tables
	table.insert( self.commands, command );
	self.commandItems[command] = CommandItem( command, false );
	
	-- refresh the list of commands
	self.listboxCommands:SetItemCount( #self.commands );
	self.listboxCommands:Refresh();
	
	-- and update the menu
	self.menu:Set( self.commands );
end

-- PERSISTENT STORAGE

function App:SerializeCommands()
	local data = {};

	-- seriazlize commands
	data.commands = {};
	for i, command in ipairs( self.commands ) do
		if i > #DefaultCommands then -- skip default commands
			local commandData = {};
			commandData.name = command:GetName();
			if command:GetPlugin() ~= nil then
				commandData.pluginName = command:GetPlugin():GetName();
			end
			commandData.shellCommand = command:GetShellCommand();
			commandData.mode = command:GetMode();
			table.insert( data.commands, commandData );
		end
	end
	
	return data;
end

function App:SerializePlugins()
	local data = {};
	
	-- serialize plugins
	data.plugins = {};
	for i, plugin in ipairs( self.plugins ) do
		data.plugins[i] = {};
		data.plugins[i].name = plugin:GetName();
		data.plugins[i].excluded = plugin:IsExcluded();
		-- data.plugins[i].autoload = pluginItem.plugin:IsAutoLoad();
	end
	
	data.showAllPlugins = self.showAllPlugins;
	
	return data;
end

function App:SerializeAutoLoad()
	local data = {};
	
	data.autoLoad = {};
	for i, plugin in ipairs( self.plugins ) do 
		if plugin:IsAutoLoad() then
			table.insert( data.autoLoad, plugin:GetName() );
		end
	end
	
	return data;
end

function App:SerializeMenu()
	local data = {};

	-- serialize menu position & status of menu button
	data.buttonPosition = {};
	data.buttonPosition.left = self.menuButton:GetLeft();
	data.buttonPosition.top = self.menuButton:GetTop();
	data.buttonScale = self.menuButton:GetScale();
	data.buttonOpacity = self.menuButton:GetOpacity();
	
	return data;
end

function App:SerializeWindow()
	local data = {};
	
	-- store visible status and position
	data.windowVisible = self:IsVisible();
	data.windowPosition = {};
	data.windowPosition.left, data.windowPosition.top = self:GetPosition();
	data.windowSize = {};
	data.windowSize.height = self:GetHeight();
	return data;
end

function App:DeserializeMenu( data )
	if data.buttonScale ~= nil then
		self.menuButton:SetScale( data.buttonScale );
	end

	-- position the menu button from stored data & set its visilibity
	if data.buttonPosition ~= nil then
		local left = Equendil.Utils.Maths.constrain( 0, data.buttonPosition.left, Turbine.UI.Display.GetWidth() - self.menuButton:GetWidth() );
		local top = Equendil.Utils.Maths.constrain( 10, data.buttonPosition.top, Turbine.UI.Display.GetHeight() - self.menuButton:GetHeight() );
		self.menuButton:SetPosition( left, top );
	end

	local opacity = 0.5;
	
	if data.buttonVisible ~= nil and not data.buttonVisible then
		opacity = 0;
	elseif data.buttonOpacity ~= nil then
		opacity = data.buttonOpacity;
	end

	self.menuButton:SetOpacity( opacity );
end

function App:DeserializeWindow( data )
	if data ~= nil then
		if data.windowSize then
			self:SetHeight( data.windowSize.height );
		end
	
		local left = Equendil.Utils.Maths.constrain( 0, data.windowPosition.left, Turbine.UI.Display.GetWidth() - self:GetWidth() );
		local top = Equendil.Utils.Maths.constrain( 0, data.windowPosition.top, Turbine.UI.Display.GetHeight() - self:GetHeight() );
		self:SetPosition( left, top );
		
		self:_Show( data.windowVisible );
	else
		self:_Show( true ); -- no settings for the window yet, we'll display it by default
	end
end
	
-- load any data we might have stored
function App:Load()
	self.persistentData = {};
	-- callbacks handling the data in defered loading

	local pluginsDataLoader = function( data ) 
		self.persistentData.plugins = data.plugins;
		if data.showAllPlugins ~= nil then
			self.showAllPlugins = data.showAllPlugins;
			self.checkboxShowHiddenPlugins:SetChecked( self.showAllPlugins );
		end
	end
	
	local commandsDataLoader = function( data )	self.persistentData.commands = data.commands; end
	local autoLoadDataLoader = function( data ) self.persistentData.autoLoad = data.autoLoad; end	

	local menuDataLoader = function( data ) self:DeserializeMenu( data ); end
	
	local dataLoader = function( loader )
		return function( data ) 
			if data ~= nil then 
				loader( data );	
			end
		end
	end
	
	-- Get the data, which we might get either as a return value or through the callback, we get it through the callback no matter what though, so it's handled there
	Equendil.Utils.PluginData.Load( Turbine.DataScope.Character, "EQ_Bootstrap_autoload", dataLoader( autoLoadDataLoader ) );
	Equendil.Utils.PluginData.Load( Turbine.DataScope.Account, "EQ_Bootstrap_commands",  dataLoader( commandsDataLoader ) );
	Equendil.Utils.PluginData.Load( Turbine.DataScope.Account, "EQ_Bootstrap_plugins",  dataLoader( pluginsDataLoader ) );
	Equendil.Utils.PluginData.Load( Turbine.DataScope.Character, "EQ_Bootstrap_menu", dataLoader( menuDataLoader ) );
	Equendil.Utils.PluginData.Load( Turbine.DataScope.Character, "EQ_Bootstrap_window", function( data ) self:DeserializeWindow( data ) end );
end

function App:SaveAutoLoad()
	function dataSaved() end -- nothing to do here
	
	local data = self:SerializeAutoLoad();
	Equendil.Utils.PluginData.Save( Turbine.DataScope.Character, "EQ_Bootstrap_autoload", data, dataSaved );
	
	self.persistentData.autoLoad = data.autoLoad;
end

function App:SavePlugins()
	function dataSaved() end -- nothing to do here
	
	local data = self:SerializePlugins();
	Equendil.Utils.PluginData.Save( Turbine.DataScope.Account, "EQ_Bootstrap_plugins", data, dataSaved );
	
	self.persistentData.plugins = data.plugins;
end

function App:SaveCommands()
	function dataSaved() end -- nothing to do here
	
	local data = self:SerializeCommands();
	Equendil.Utils.PluginData.Save( Turbine.DataScope.Account, "EQ_Bootstrap_commands", data, dataSaved );
	
	self.persistentData.commands = data.commands;
end

function App:SaveMenu()
	function dataSaved() end -- nothing to do here
	local menuData = self:SerializeMenu();
	Equendil.Utils.PluginData.Save( Turbine.DataScope.Character, "EQ_Bootstrap_menu", menuData, dataSaved );
end

function App:SaveWindow()
	function dataSaved() end 
	
	local windowData = self:SerializeWindow();
	Equendil.Utils.PluginData.Save( Turbine.DataScope.Character, "EQ_Bootstrap_window", windowData, dataSaved );
end